import os
import pandas as pd
from pyarrow.pandas_compat import _index_level_name

from azureml.studio.core.utils.fileutils import ExecuteInDirectory
from azureml.studio.core.logger import logger


_PARQUET_ENGINE = 'pyarrow'


def _get_arrow_index_names(df: pd.DataFrame):
    """Get index names which will be generated by pyarrow."""
    n = len(getattr(df.index, 'levels', [df.index]))
    return [_index_level_name(df.index.get_level_values(i), i, df.columns) for i in range(n)]


def _remove_arrow_bug_columns(df):
    # Due to a bug in pyarrow, if the name of a column is the same as the index generated by arrow,
    # an assertion error will be raised when trying reading the parquet file.
    # See https://github.com/apache/arrow/blob/maint-0.12.x/python/pyarrow/pandas_compat.py#L590
    # Repro steps:
    # pd.DataFrame({'__index_level_0__': [1]}).to_parquet(f)
    # pd.read_parquet(f)

    # To avoid such problem, we have to find the invalid columns and drop them before saving parquet files.
    indexes = _get_arrow_index_names(df)
    invalid_columns = [index for index in indexes if index in df.columns]
    if invalid_columns:
        df = df.drop(columns=invalid_columns)
        logger.warning(f"Columns are dropped because their names are invalid to save parquet file. "
                       f"The dropped columns are: {','.join(invalid_columns)}")
    return df


def data_frame_to_parquet(df: pd.DataFrame, parquet_path):
    preserve_index = True
    # If the index is the default index (0, 1, ... n-1) which will be set when initializing a DataFrame,
    # we do not preserve the index to reduce cost and avoid bugs.
    # The operation is O(1) since both indexes are RangeIndex, the operation doesn't introduce much cost.
    if isinstance(df.index, pd.RangeIndex):
        default_index = pd.RangeIndex(0, df.shape[0])
        if df.index.equals(default_index):
            preserve_index = False

    # If we preserve the index, exception will be caused if some column names are invalid due to arrow's bug.
    # We need to remove such columns before writing to parquet.
    if preserve_index:
        df = _remove_arrow_bug_columns(df)

    # Use chdir to dest_directory to work around the path issue on windows when writing parquet file
    with ExecuteInDirectory(parquet_path) as parquet_file_name:
        # Use len(df.columns)==0 instead of df.empty because df.empty is true when the df has column names but no row.
        # In this case, the column names should be stored.
        if df is not None and not len(df.columns) == 0:
            df.to_parquet(fname=parquet_file_name, engine=_PARQUET_ENGINE,
                          index=preserve_index,
                          # Fix bug 484283: pyarrow raises error due to the precision problem of timestamps.
                          # Add this option so the nanoseconds will be converted to milliseconds
                          allow_truncated_timestamps=True,
                          )


def data_frame_from_parquet(parquet_path):
    """
    Read pandas DataFrame from parquet file
    :param parquet_path: str
    :return: pd.DataFrame
    """
    # Use chdir to dest_directory to work around the path issue on windows when writing parquet file
    with ExecuteInDirectory(parquet_path) as parquet_file_name:
        if os.path.exists(parquet_file_name):
            return pd.read_parquet(parquet_file_name, engine=_PARQUET_ENGINE)
        else:
            return pd.DataFrame()


def data_frame_to_csv(df: pd.DataFrame, csv_path, has_header=True):
    df.to_csv(path_or_buf=csv_path, header=has_header, index=False)
